/* hsen.cpp - 03 Apr 09
   Hosts3D - 3D Real-Time Network Monitor
   Copyright (c) 2006-2009  Del Castle

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details. */

#include <arpa/inet.h>  //inet_addr()
#include <arpa/nameser.h>  //HEADER
#include <netinet/if_ether.h>  //ether_arp
#include <netinet/ip.h>  //iphdr
#include <netinet/tcp.h>  //tcphdr
#include <netinet/udp.h>  //udphdr
#include <pcap.h>
#include <signal.h>  //signal()
#include <stdlib.h>  //atoi()
#include <string.h>  //memcpy()
#include <sys/stat.h>  //umask()
#include <syslog.h>  //syslog()
#include <unistd.h>  //usleep(), fork(), setsid(), close()

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "proto.h"
#include "misc.h"

const int pesz = sizeof(pkex_type), dnsz = sizeof(pdns_type), MAXEPS = 331;  //max ethernet packet size PPP 2 ETH 14 IP 20 UDP 8 DNS HEAD 12 DNS NAME 255 DNS TAIL 20
const unsigned int ehsz = sizeof(ether_header), ahsz = ehsz + sizeof(ether_arp), ihsz = ehsz + sizeof(iphdr), thsz = ihsz + sizeof(tcphdr), uhsz = ihsz + sizeof(udphdr)
  , dhsz = uhsz + sizeof(HEADER), dtsz = dhsz + sizeof(dns_tail);

bool rpy = false, ppp;  //linux cooked sockets (ppp) present
int hsock;  //hosts3d udp socket
sockaddr_in hadr;  //destination ip/broadcast address
unsigned int hasz = sizeof(hadr);
unsigned long long nowtm, reltm, rpoff = 0;  //pcap replay time offset
pcap_t *pcap;

void hsenStop(int sig)
{
  syslog(LOG_INFO, "stopping...\n");
  pcap_close(pcap);
}

//pcap_loop
void pktProcess(u_char *u, const struct pcap_pkthdr *hdr, const u_char *pkt)
{
  if (rpy)  //pcap replay
  {
    nowtm = microTime(0);
    if (!rpoff) rpoff = nowtm - microTime(&hdr->ts);
    reltm = microTime(&hdr->ts) + rpoff;
    if (reltm > nowtm) usleep(reltm - nowtm);
  }
  pkif_type pkif = {0, 0, {0}, {0}, atoi((const char *)u), 250};  //protocol 250 unassigned used to identify unknown IP packet
  pkex_type pkex = {85, 0, 0, hdr->len, {0, 0, 0, 0, 0, 0}};  //85 used to identify packet info
  pkex.pk = pkif;
  bool known = false;
  unsigned int clen = hdr->caplen;
  if (clen < ehsz) return;  //malformed packet check
  if (ppp)
  {
    pkt += 2;
    clen -= 2;
    pkex.sz -= 2;
  }
  ether_header *ehdr = (ether_header *)pkt;
  memcpy(&pkex.srcmc, &ehdr->ether_shost, 6);
  if (ntohs(ehdr->ether_type) == ETHERTYPE_IP)
  {
    if (clen < ihsz) return;  //malformed packet check
    iphdr *ihdr = (iphdr *)(pkt + ehsz);
    unsigned short frag = ntohs(ihdr->frag_off) << 3;
    if (frag) return;  //fragmented packet check
    pkex.sz -= ehsz;
    pkex.pk.pr = ihdr->protocol;
    pkex.pk.srcip.s_addr = ihdr->saddr;
    pkex.pk.dstip.s_addr = ihdr->daddr;
    if (pkex.pk.pr == IPPROTO_TCP)
    {
      if (clen < thsz) return;  //malformed packet check
      tcphdr *thdr = (tcphdr *)(pkt + ihsz);
      pkex.pk.srcpt = ntohs(thdr->source);
      pkex.pk.dstpt = ntohs(thdr->dest);
      pkex.syn = thdr->syn;
      pkex.ack = thdr->ack;
    }
    else if (pkex.pk.pr == IPPROTO_UDP)
    {
      if (clen < uhsz) return;  //malformed packet check
      udphdr *uhdr = (udphdr *)(pkt + ihsz);
      if (hadr.sin_port == uhdr->dest) return;  //ignore hsen to hosts3d packets
      pkex.pk.srcpt = ntohs(uhdr->source);
      pkex.pk.dstpt = ntohs(uhdr->dest);
      udp_type udpt = udpTrack(&pkex.pk);
      if (udpt == src) pkex.ack = 1;  //ack used to identify service
      if ((pkex.pk.srcpt == 53) && (clen > dtsz))  //port 53 DNS, malformed packet check
      {
        HEADER *dhdr = (HEADER *)(pkt + uhsz);
        if ((dhdr->qr == 1) && !dhdr->opcode && !dhdr->rcode && (ntohs(dhdr->qdcount) == 1) && (ntohs(dhdr->ancount) == 1))
        {
          pdns_type pdns = {42, "", {0}};  //42 used to identify DNS info
          char *dnsdata = (char *)(pkt + dhsz);
          addrGet(dnsdata, pdns.htnm, &pdns.dnsip, clen - dtsz);
          if (pdns.dnsip.s_addr)
          {
            if (sendto(hsock, &pdns, dnsz, 0, (sockaddr *)&hadr, hasz) != dnsz)  //send dns info to hosts3d
            {
              syslog(LOG_ERR, "socket send failed (dns info)\n");
              hsenStop(0);
            }
          }
        }
      }
    }
    known = true;
  }
  else if (ntohs(ehdr->ether_type) == ETHERTYPE_ARP)
  {
    if (clen < ahsz) return;  //malformed packet check
    ether_arp *ahdr = (ether_arp *)(pkt + ehsz);
    pkex.sz = 0;
    pkex.pk.pr = 249;  //protocol 249 unassigned used to identify ARP packet
    memcpy(&pkex.pk.srcip, &ahdr->arp_spa, 4);
    memcpy(&pkex.pk.dstip, &ahdr->arp_tpa, 4);
    known = true;
  }
  if (known)
  {
    if (sendto(hsock, &pkex, pesz, 0, (sockaddr *)&hadr, hasz) != pesz)  //send packet info to hosts3d
    {
      syslog(LOG_ERR, "socket send failed (pkt info)\n");
      hsenStop(0);
    }
  }
}

int main(int argc, char *argv[])
{
  pid_t pid = fork();
  if (pid == -1)
  {
    fprintf(stderr, "hsen error: fork() failed\n");
    return 1;
  }
  if (pid > 0) return 0;
  pid_t sid = setsid();
  if (sid == -1)
  {
    fprintf(stderr, "hsen error: setsid() failed\n");
    return 1;
  }
  umask(0);
  if (!((argc == 3) || (argc == 4)))
  {
    fprintf(stderr, "hsen 0.98 usage: %s <id> <interface/file> [<destination>]\n", argv[0]);
    return 1;
  }
  if ((atoi(argv[1]) < 1) || (atoi(argv[1]) > 255))
  {
    fprintf(stderr, "hsen error: id out of range [1 - 255]\n");
    return 1;
  }
  hadr.sin_addr.s_addr = (argc == 4 ? inet_addr(argv[3]) : inet_addr("127.0.0.1"));  //hosts3d at localhost
  if (!hadr.sin_addr.s_addr)
  {
    fprintf(stderr, "hsen error: invalid destination ip/broadcast address\n");
    return 1;
  }
  syslog(LOG_INFO, "started\n");
  hadr.sin_family = AF_INET;
  hadr.sin_port = htons(10111);  //port 10111 unassigned, used for hsen-to-hosts3d comms
  if ((hsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
  {
    syslog(LOG_ERR, "socket create failed\n");
    fprintf(stderr, "hsen error: socket create failed\n");
    return 1;
  }
  int sopt = 1;
  if (setsockopt(hsock, SOL_SOCKET, SO_NO_CHECK, &sopt, sizeof(sopt)) == -1)  //disable udp checksum
  {
    syslog(LOG_ERR, "socket opt failed- checksum\n");
    fprintf(stderr, "hsen error: socket opt failed- checksum\n");
    close(hsock);
    return 1;
  }
  if (setsockopt(hsock, SOL_SOCKET, SO_BROADCAST, &sopt, sizeof(sopt)) == -1)  //enable udp broadcasts
  {
    syslog(LOG_ERR, "socket opt failed- broadcast\n");
    fprintf(stderr, "hsen error: socket opt failed- broadcast\n");
    close(hsock);
    return 1;
  }
  char errbuf[PCAP_ERRBUF_SIZE];
  if (!(pcap = pcap_open_offline(argv[2], errbuf)))  //test for file or interface input
  {
    syslog(LOG_INFO, "pcap_open_offline() failed- %s, trying pcap_open_live()\n", errbuf);
    if (!(pcap = pcap_open_live(argv[2], MAXEPS, 1, 0, errbuf)))
    {
      syslog(LOG_ERR, "pcap_open_live() failed- %s, r u root, iface up?\n", errbuf);
      fprintf(stderr, "hsen error: pcap_open_live() failed- %s, r u root, iface up?\n", errbuf);
      close(hsock);
      return 1;
    }
  }
  else if (strcmp(argv[2], "-")) rpy = true;
  ppp = (pcap_datalink(pcap) == DLT_LINUX_SLL);  //linux cooked sockets (ppp)
  signal(SIGTERM, hsenStop);  //capture kill
  pcap_loop(pcap, -1, pktProcess, (u_char *)argv[1]);
  close(hsock);
  utrkDestroy();  //destroy UDP packet tracking data
  syslog(LOG_INFO, "stopped\n");
  return 0;
}
